ANTLR_VERSION = '4.13.2'
ANTLR_JAR_NAME = f'antlr-{ANTLR_VERSION}-complete.jar'
ANTLR_JAR_URL = f'https://www.antlr.org/download/{ANTLR_JAR_NAME}'
ANTLR_RUNTIME_PACKAGE = 'antlr4-python3-runtime'
ANTLR_GITHUB_REPO = 'antlr/antlr4'

ANTLR_RUNTIME_VENDOR = 'omextra.text.antlr._runtime._all'
